from __future__ import annotations

from asyncio.tasks import Task
from typing import Callable, Dict, Tuple, cast

from . import current_manager
from .types import TRunArgs, TTaskParams
from .utils import to_coroutinefn


class TaskWrapper:
    """Wrap a given function into a Task object."""

    __slots__ = ("_manager", "_fn", "_params", "_timer")

    def __init__(self, manager: Donald, fn: Callable, params: TTaskParams):
        if "<locals>" in fn.__qualname__:
            raise ValueError("Cannot use local functions as tasks")

        self._manager = manager
        self._fn = to_coroutinefn(fn)
        self._params: TTaskParams = params
        self._timer = None

    def __repr__(self):
        return f"<TaskWrapper {self._fn.__qualname__}>"

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)

    def import_path(self) -> str:
        return f"{self._fn.__module__}.{self._fn.__qualname__}"

    def submit(self, *args, **kwargs):
        return self.apply_submit(*args, kwargs=kwargs)

    def apply_submit(self, *args, kwargs: Dict = {}, **params):
        task_params = cast(TTaskParams, dict(self._params, **params))
        res = TaskRun(self.import_path(), args, kwargs, task_params)
        return self._manager.submit(res)


class TaskRun:
    """Wrap a given function into a Task object."""

    __slots__ = ("data", "retries")

    def __init__(
        self,
        path: str,
        args: Tuple,
        kwargs: Dict,
        params: TTaskParams,
    ):
        if params.get("bind"):
            args = (self,) + args

        self.retries = 0
        self.data: TRunArgs = (path, args, kwargs, params)

    def retry(self: TaskRun):
        manager: Donald = current_manager.value
        self.retries += 1
        return manager.submit(self)

    def __repr__(self):
        path = self.data[0]
        return f"<TaskRun {path}>"


from .manager import Donald



"""Tasks' manager for Asyncio."""

__version__ = "0.31.9"
__project__ = "Donald"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "BSD"

import logging

logger = logging.getLogger(__name__)

from threading import local

current_manager = local()
current_manager.value = None

from .manager import Donald  # noqa
from .tasks import TaskRun, TaskWrapper  # noqa



import asyncio
import random
from pickle import dumps, loads
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, Mapping, Type
from urllib.parse import urlparse
from uuid import uuid4

from . import logger
from .types import TBackendType, TRunArgs

if TYPE_CHECKING:
    from aioamqp import Channel
    from redis.asyncio import Redis


class BaseBackend:

    type: TBackendType
    defaults: Mapping = {}

    def __init__(self, params: Mapping):
        self.is_connected = False
        self.params = dict(self.defaults, **params)
        self.__backend__: Any = None

    async def connect(self):
        logger.info("Connecting to Tasks Backend: %s", self.type)
        await self._connect()
        self.is_connected = True
        logger.info("Connected to Tasks Backend: %s", self.type)

    async def disconnect(self):
        logger.info("Disconnecting from Tasks Backend: %s", self.type)
        self.is_connected = False
        await self._disconnect()
        logger.info("Disconnected from Tasks Backend: %s", self.type)

    def submit(self, data: TRunArgs):
        if not self.is_connected:
            raise RuntimeError("Tasks Backend is not connected")
        _data = dumps(data)
        return self._submit(_data)

    async def subscribe(self) -> AsyncIterator[TRunArgs]:
        raise NotImplementedError

    async def _connect(self):
        raise NotImplementedError

    async def _disconnect(self):
        pass

    async def _submit(self, data):
        raise NotImplementedError


class MemoryBackend(BaseBackend):
    type: TBackendType = "memory"

    @property
    def rx(self) -> asyncio.Queue:
        if self.__backend__ is None:
            raise RuntimeError("Tasks Backend is not connected")

        return self.__backend__

    async def _connect(self):
        self.__backend__ = asyncio.Queue()

    async def _submit(self, data):
        self.rx.put_nowait(data)
        return True

    async def subscribe(self):
        await self.connect()

        async def iter_tasks():
            while self.is_connected:
                msg = await self.rx.get()
                if msg is None:
                    continue

                yield loads(msg)
                await asyncio.sleep(0)

        return iter_tasks()


class RedisBackend(BaseBackend):
    type: TBackendType = "redis"
    defaults = {
        "url": "redis://localhost:6379/0",
        "channel": "donald",
    }

    @property
    def redis(self) -> "Redis":
        if self.__backend__ is None:
            raise RuntimeError("Tasks Backend is not connected")

        return self.__backend__

    async def _connect(self):
        from redis.asyncio import from_url

        self.__backend__ = from_url(self.params["url"])

    async def _disconnect(self):
        await self.redis.close()

    async def _submit(self, data):
        return await self.redis.publish(f"{self.params['channel']}:1", data)

    async def subscribe(self):
        pubsub = self.redis.pubsub(ignore_subscribe_messages=True)
        await pubsub.subscribe(f"{self.params['channel']}:1")

        async def iter_tasks():
            while self.is_connected:
                msg = await pubsub.get_message()
                if msg is None:
                    await asyncio.sleep(1e-2)
                    continue

                yield loads(msg["data"])

        return iter_tasks()


class AMQPBackend(BaseBackend):
    type: TBackendType = "amqp"
    defaults = {
        "url": "amqp://guest:guest@localhost:5672/",
        "queue": "donald",
        "exchange": "",
    }

    def __init__(self, params: Mapping):
        super().__init__(params)
        self.__protocol__ = None
        self.__channel__ = None

    @property
    def channel(self) -> "Channel":
        if self.__backend__ is None:
            raise RuntimeError("Tasks Backend is not connected")

        return self.__channel__

    async def _connect(self):
        from aioamqp import connect

        url = urlparse(self.params["url"])
        host, port, vhost, user, password = (
            url.hostname,
            url.port,
            url.path,
            url.username,
            url.password,
        )

        try:
            self.__backend__, self.__protocol__ = await connect(
                host=host,
                port=port,
                virtualhost=vhost,
                login=user,
                password=password,
                on_error=self.on_error,
                **self.params,
            )
        except OSError as exc:
            logger.exception("Failed to connect to AMQP")
            return asyncio.create_task(self.on_error(exc))

        self.__channel__ = await self.__protocol__.channel()
        await self.__channel__.queue_declare(
            queue_name=self.params["queue"], durable=True
        )
        await self.__channel__.basic_qos(
            prefetch_count=1, prefetch_size=0, connection_global=False
        )

    async def _disconnect(self):
        await self.__protocol__.close(no_wait=True)
        self.__backend__.close()

    reconnecting = None

    async def on_error(self, exc):
        if not self.is_connected:
            return

        self.is_connected = False
        if self.reconnecting is None:
            self.reconnecting = asyncio.Condition()

        if self.reconnecting.locked():
            return

        async with self.reconnecting:
            logger.exception("AMQP Connection Error, reconnecting in 5 seconds")
            await asyncio.sleep(5 + random.random())
            try:
                await self.connect()
            except Exception:
                pass

    async def _submit(self, data):
        await self.channel.basic_publish(
            payload=data,
            exchange_name=self.params["exchange"],
            routing_key=self.params["queue"],
            properties=dict(delivery_mode=2, message_id=f"{uuid4()}"),
        )
        return True

    async def subscribe(self):
        rx = asyncio.Queue()

        async def consumer(channel, body, envelope, properties):
            rx.put_nowait(body)
            await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

        await self.__channel__.basic_consume(consumer, self.params["queue"])

        async def iter_tasks():
            while self.is_connected:
                msg = await rx.get()
                if msg is None:
                    continue

                yield loads(msg)

        return iter_tasks()


BACKENDS: Dict[TBackendType, Type[BaseBackend]] = {
    "memory": MemoryBackend,
    "redis": RedisBackend,
    "amqp": AMQPBackend,
}



from __future__ import annotations

from asyncio.tasks import create_task
from logging.config import dictConfig
from typing import Callable, Optional, cast, overload

from . import current_manager, logger
from .backend import BACKENDS, BaseBackend
from .types import TManagerParams, TTaskParams, TVWorkerOnErrFn, TVWorkerOnFn, TWorkerParams
from .worker import Worker


class Donald:
    """Manage tasks and workers."""

    defaults: TManagerParams = {
        "log_level": "INFO",
        "log_config": None,
        "backend": "memory",
        "backend_params": {},
        "worker_params": cast(TWorkerParams, {}),
    }

    _backend: BaseBackend

    def __init__(self: Donald, **params):
        self._params: TManagerParams = self.defaults
        self.is_started = False
        self.setup(**params)
        self.scheduler = Scheduler()
        current_manager.value = self

    def __repr__(self):
        return f"<Donald {self._params['backend']}>"

    def setup(self: Donald, **params):
        """Setup the manager."""
        if self.is_started:
            raise RuntimeError("Manager is already started")

        self._params = cast(TManagerParams, dict(self._params, **params))
        self._backend = BACKENDS[self._params["backend"]](
            self._params["backend_params"]
        )

        logger.setLevel(self._params["log_level"].upper())
        if self._params["log_config"]:
            dictConfig(self._params["log_config"])

    async def start(self):
        """Start the manager."""
        logger.info("Starting tasks manager")
        await self._backend.connect()
        self.is_started = True
        logger.info("Tasks manager started")

    async def stop(self):
        logger.info("Stopping tasks manager")
        await self._backend.disconnect()
        self.is_started = False
        logger.info("Tasks manager stopped")

    def create_worker(self: Donald, **params):
        """Create a worker."""
        worker_params = cast(
            TWorkerParams, dict(self._params["worker_params"], **params)
        )
        return Worker(self._backend, worker_params)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    def schedule(self, interval: TInterval):
        """Schedule a task to the backend."""
        return self.scheduler.schedule(interval)

    @overload
    def task(self, fn: Callable) -> TaskWrapper:
        ...

    @overload
    def task(self, **params) -> Callable[[Callable], TaskWrapper]:
        ...

    def task(self, fn: Optional[Callable] = None, **params):
        """Decorator to wrap a function into a Task object."""

        def wrapper(fn: Callable) -> TaskWrapper:
            return TaskWrapper(self, fn, cast(TTaskParams, params))

        if fn is None:
            return wrapper

        return wrapper(fn)

    def submit(self, run: TaskRun):
        """Submit a task to the backend."""
        if not self.is_started:
            raise RuntimeError("Manager is not started")

        return create_task(self._backend.submit(run.data))

    def on_start(self, fn: TVWorkerOnFn) -> TVWorkerOnFn:
        """Register a function to be called on worker start."""
        self._params["worker_params"]["on_start"] = fn
        return fn

    def on_stop(self, fn: TVWorkerOnFn) -> TVWorkerOnFn:
        """Register a function to be called on worker stop."""
        self._params["worker_params"]["on_stop"] = fn
        return fn

    def on_error(self, fn: TVWorkerOnErrFn) -> TVWorkerOnErrFn:
        """Register a function to be called on worker error."""
        self._params["worker_params"]["on_error"] = fn
        return fn


from .scheduler import Scheduler, TInterval
from .tasks import TaskRun, TaskWrapper



from __future__ import annotations

from numbers import Number
from typing import (Any, Awaitable, Callable, Coroutine, Dict, Literal, Optional, Tuple, TypedDict,
                    TypeVar)

TWorkerOnFn = Callable[[], Coroutine]
TVWorkerOnFn = TypeVar("TVWorkerOnFn", bound=TWorkerOnFn)
TWorkerOnErrFn = Callable[[BaseException], Coroutine]
TVWorkerOnErrFn = TypeVar("TVWorkerOnErrFn", bound=TWorkerOnErrFn)

TV = TypeVar("TV", bound=Any)
TAsyncFn = Callable[..., Coroutine]
TVAsyncFn = TypeVar("TVAsyncFn", bound=TAsyncFn)

TBackendType = Literal["memory", "redis", "amqp"]

TTaskParams = TypedDict(
    "TTaskParams",
    {
        "bind": Optional[bool],
        "delay": Optional[Number],
        "timeout": Optional[Number],
    },
)

TRunArgs = Tuple[str, Tuple, Dict[str, Any], TTaskParams]


class TWorkerParams(TypedDict):
    max_tasks: Optional[int]
    task_defaults: Optional[TTaskParams]
    on_start: Optional[TWorkerOnFn]
    on_stop: Optional[TWorkerOnFn]
    on_error: Optional[TWorkerOnErrFn]
    show_banner: Optional[bool]
    # "timeout": Number,
    # "task_retry": int,
    # "task_retry_delay": Number,


class TManagerParams(TypedDict):
    backend: TBackendType
    backend_params: Dict
    log_config: Optional[Dict]
    log_level: str
    worker_params: TWorkerParams
